/*
 * Copyright (c) 2009, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * The names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#ifndef SPTH_CL
#include <QCoreApplication>
#endif /* SPTH_CL */

#include "game.h"
#include "dict.h"
#include "util.h"

#ifdef WIN32
/* mingw doesn't have random(3) */
# define random rand
# define srandom srand
#endif /* WIN32 */

namespace Spellathon
{
	Game::Game(Dict *dct, bool plural_sieve) 
		: dict_(dct), found_(false), plural_sieve_(plural_sieve)
	{
		srandom(time(0));
		words_.reserve(128);
		wordPtrs.reserve(128); 
	}

	Game::~Game()
	{
	}

	void 
	Game::setPluralSieve(bool e)
	{ 
		plural_sieve_ = e;
	}

	bool
	Game::play(bool precomp, int minsol, int maxsol)
	{
		if (precomp) 
			return _play_precomp(minsol, maxsol);
		else
			return _play(minsol, maxsol);
	}

	bool
	Game::_play(int minsol, int maxsol)
	{
		char		sig[MAX_LETRS+1];
		unsigned int 	(*sidx)[2];
		enum	{IDX, RAND};

	    sidx = (unsigned int (*)[2])malloc(2 * dict_->nent7() * sizeof(int));
		if (sidx == NULL) 
			spth_error(true, "Out of memory");

		/* shuffle the indexes. Knuth shuffle? */
		int j = dict_->off7();
		if (j < 0)
			goto notfound;

        for (int i = 0; i < dict_->nent7(); i++) {
                sidx[i][IDX] = j++;
                sidx[i][RAND] = random();
        }
	
        qsort((void*)sidx, dict_->nent7(), 2 * sizeof(int), sidxcmp);
	
		unsigned int cidx[MAX_LETRS][2];
        for (int i = 0; i < MAX_LETRS; i++) {
                cidx[i][IDX] = i;
                cidx[i][RAND] = random();
        }
	
        qsort((void*)cidx, MAX_LETRS, 2 * sizeof(int), sidxcmp);
	
		/*
		 * find a random seven letter signature from the dict and solve
		 * it and any of the six different central letter combinations.
		 * consider the word playable if MIN_SOL <= solutions <= MAX_SOL
		 *
		 * this technique works well. it's fast, sufficiently
		 * random and covers all the signatures.
		 */
		for (int i = 0; i < dict_->nent7(); i++) {

			const char *dsig = (*dict_).dictarr[sidx[i][IDX]].sign;

			int j = 0;
			while (j < MAX_LETRS) {
				sig[j] = dsig[cidx[j][IDX]]; 
				j++;
			}
			sig[j] = 0;

			int k = MAX_LETRS - 2;
			do {
				if (solve(sig, minsol, maxsol) != -1) {
					/* found suitable sign/game */
					free(sidx);
					sig_ = sig;
					return found_ = true;
				}

				if (k < 0)
					break;

				/* try a different central letter */	
				swapch(&sig[MAX_LETRS-1], &sig[k--]);
	
			} while (1);

			/* 
			 * this can take some time in pathological cases 
			 * so we need to keep the gui responsive 
			 */	
			if (!(i & ((1<<8) - 1)) && i > 1024) {
#ifdef SPTH_CL
				spinner();
#else
				QCoreApplication::processEvents();
#endif /* SPTH_CL */
			}
		}
notfound:
		/* Uh Oh */	
		free(sidx);
		return found_ = false;
	}

	bool
	Game::_play_precomp(int minsol, int maxsol)
	{
		int i = 0, s = -1, e = -1;

		while (i < dict_->precomp_nent()) {
			if (dict_->precomparr[i].nsol >= minsol) {
				s = i;
				break;
			}
			i++;
		}

		while (i < dict_->precomp_nent()) {
			if (dict_->precomparr[i].nsol > maxsol) {
				break;
			}
			i++;
		}

		e = i;

		if (s < 0 || s >= e) {
			return found_ = false;
		} else {
			int r = randnum(s, e);
			sig_ = dict_->precomparr[r].sign;
			solve(sig_.c_str());
			return found_ = true;
		}
	}

	bool
	Game::solve(string letters)
	{
		solve(letters.c_str());
		return found_ = nwords() > 0;
	}

	int 
	Game::sidxcmp(const void *a, const void *b)
	{
	        const unsigned int (*c)[2] = (const unsigned int (*)[2])a;
	        const unsigned int (*d)[2] = (const unsigned int (*)[2])b;
	
	        if ((*c)[1] == (*d)[1])
	                return 0;
	        if ((*c)[1] < (*d)[1])
	                return -1;
	        else
	                return 1;
	}

	/* performance critical routine */
	int
	Game::solve(const char* letrs, int min, int max)
	{
		char	signarr[NCOMBINATIONS][MAX_LETRS+1];
	
		int c = signGen(letrs, signarr);
	
		wordPtrs.clear();
		/* 
		 * for all the sign. generated by signgen() find matching 
		 * sign (and thus words) from the dictionary
		 */
		if (plural_sieve_) {
			/* remove plurals */
			bool has7 = false;
			vector<char *> w;
			w.reserve(64);
			for (int k = 0; k < c; k++) {
				w.clear();
				dict_->matchSign(signarr[k], w);
				int i = (int)w.size();
				while (--i >= 0) {
					int slen = strlen(w[i]);
						int slen2 = slen - 1;
						/* .*s */
						if (w[i][slen2] == 's') {
							/* promiscuous pointer manipulation below */
							w[i][slen2] = 0;	
							if (dict_->matchWord(w[i])) {
								w[i][slen2] = 's';
								continue;
							} 
							w[i][slen2] = 's';
							/* .*es */
							int slen3 = slen2 - 1;
							if (w[i][slen3] == 'e') {
								w[i][slen3] = 0;
								w[i][slen2] = 0;
								if (dict_->matchWord(w[i])) {
									w[i][slen3] = 'e';
									w[i][slen2] = 's';
									continue;
								} 
								w[i][slen3] = 'e';
								w[i][slen2] = 's';
							}
							/* .*ies */
							int slen4 = slen3 - 1;
							if (w[i][slen4] == 'i') {
								w[i][slen4] = 'y';
								w[i][slen3] = 0;
								w[i][slen2] = 0;
								if (dict_->matchWord(w[i])) {
									w[i][slen4] = 'i';
									w[i][slen3] = 'e';
									w[i][slen2] = 's';
									continue;
								} 
								w[i][slen4] = 'i';
								w[i][slen3] = 'e';
								w[i][slen2] = 's';
							}
						}
					has7 = (slen == 7);
					wordPtrs.push_back(w[i]); 
				}
				if ((int)wordPtrs.size() > max)
					return -1;
			}
			if (!has7) 
				return -1;
		} else { 
			for (int i = 0; i < c; i++) {
				dict_->matchSign(signarr[i], wordPtrs);
				if ((int)wordPtrs.size() > max)
					return -1;
			}
		}
	
		if ((int)wordPtrs.size() < min)
			return -1;

		return wordPtrs.size();
	}
	
	/* 
	 * find all the letter combinations of strlen 4, 5, 6 and 7.
	 * the letters can be duplicates so that needs to be handled.
	 * since this a fixed problem i have used a table lookup.
	 */
	/* performance critical routine */
	int
	Game::signGen(const char *letters, char signarr[][MAX_LETRS+1])
	{
		int		i, j, k;
		const uint8_t	smask[NCOMBINATIONS] = { 
			/* C(6,3) */
			0x7,  0xb,  0xd,  0xe,  0x13, 0x15, 
			0x16, 0x19, 0x1a, 0x1c, 0x23, 0x25, 
			0x26, 0x29, 0x2a, 0x2c, 0x31, 0x32, 
			0x34, 0x38,
			/* C(6,4) */
			0xf,  0x17, 0x1b, 0x1d, 0x1e, 0x27, 
			0x2b, 0x2d, 0x2e, 0x33, 0x35, 0x36, 
			0x39, 0x3a, 0x3c,
			/* C(6,5) */
			0x1f, 0x2f, 0x37, 0x3b, 0x3d, 0x3e,
			/* C(6,6) */
			0x3f 
		}; 
		int8_t coff[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
	
		memset(signarr, 0, NCOMBINATIONS * (MAX_LETRS+1));

		int n = 0;

		char letrlookup[64];
		letrlookup[0] = 0;
		letrlookup[0x1] = letters[0];
		letrlookup[0x2] = letters[1];
		letrlookup[0x4] = letters[2];
		letrlookup[0x8] = letters[3];
		letrlookup[0x10] = letters[4];
		letrlookup[0x20] = letters[5];
		/* 
		 * the central letter is present in all the words so
		 * find the combinations of the remaining 6 letters 
		 */
		for (i = 0; i < NCOMBINATIONS; i++) {
			/* generate a letter combination from the bitmask */
			j = 0;
			signarr[n][j] = letrlookup[smask[i]&1	]; j +=  smask[i]&1;
			signarr[n][j] = letrlookup[smask[i]&0x2	]; j += (smask[i]&0x2)>>1;
			signarr[n][j] = letrlookup[smask[i]&0x4	]; j += (smask[i]&0x4)>>2;
			signarr[n][j] = letrlookup[smask[i]&0x8	]; j += (smask[i]&0x8)>>3;
			signarr[n][j] = letrlookup[smask[i]&0x10]; j += (smask[i]&0x10)>>4;
			signarr[n][j] = letrlookup[smask[i]&0x20]; j += (smask[i]&0x20)>>5;
	
			/* add the central letter */
			signarr[n][j++] = letters[MAX_LETRS-1];
	
			/* sign it */
			sign(signarr[n], j);
	
			if (coff[j] == -1)
			   coff[j] = n;

			/* ignore duplicate comb. in the case where letters 
			 * are duplicates */
			k = n-1;
			while (k >= coff[j]) {
				if (strcmp_em(signarr[n], signarr[k])) {
					n--;	
					break;
				}
				k--;
			}
			n++;
		}
	
		return n;	
	}

	const vector<string> & 
	Game::words() 
	{ 
		words_.clear();

		for (int i = 0; i < (int)wordPtrs.size(); i++) 
			words_.push_back(wordPtrs[i]); 

		return words_;
	}

}

